#!@@PERL@@
#
# machine.cgi						Author: Rich West
#							rwest@wesmo.com
#
# Acting as a GUI front end to all of the core system files for networked
# machines, this CGI script provides an interface for admins to modify
# (add/delete) information from the MySQL machine databases.
#
# It also provides a secure and simplified way of management over pushing
# out that information to the respective servers on the network to make those
# changes reflect across all systems.
#
# $Header: /home/rwest/work/MID/src/RCS/machine.cgi,v 1.13 2001/04/03 17:23:42 rwest Exp rwest $
################################################################################

################################################################################
#
# The name of this script.
#
################################################################################
$VERSION = "@@VERSION@@";
$my_name=`basename $0`;
chop($my_name);

################################################################################
#
# The modules we require.
#
################################################################################
use Mysql;
use CGI::Cookie;

&main();

sub db_settings
{
################################################################################
#
# Our Database Information
#
################################################################################
$host = "@@DBHOST@@";
$dbname = "@@DB@@";
$user = "@@DBUSER@@";
$pass = "@@DBPASS@@";
}
 
################################################################################
#
# Grab all of our defaults from the database 'conf' table.
#
################################################################################
sub defaults
{
   local($qs, @row);
   $qs = "SELECT domainname,named,yp,tmp,start_range,end_range,dhcpd,SERIAL FROM conf";
   $sth = $dbh->Query($qs)
      or &notification("[defaults] Query had some problem: $Mysql::db_errstr\n$qs\n");
   @row=$sth->FetchRow;
   $DOMAINNAME = $row[0];
   @RANGE=($row[4],$row[5]);
   $NAMED_DIR=$row[1];
   $NIS_DIR=$row[2];
   $TMP=$row[3];
   $DHCPD_DIR=$row[6];
   $SERIAL = $row[7];

#
# These are the fields that we do not want to show up in the list of available
# ways to sort and we do not want them showing up when displaying the database
# contents since it would just make it waaaaay to messy.
#
   $noshow{"OS"} = 1;
   $noshow{"Comment"} = 1;
   $noshow{"Serial_Number"} = 1;
   $noshow{"Processor"} = 1;
   $noshow{"Processor_Speed"} = 1;
   $noshow{"Memory"} = 1;
}

sub format_styles
{
################################################################################
#
# Our Format Styles
#
################################################################################
format DNS_A =
@<<<<<<<<<<<<<<<<<<<<	IN	A	@<<<<<<<<<<<<<<<
$name,$ip
.

format DNS_C =
@<<<<<<<<<<<<<<<<<<<<	IN	CNAME	@<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$alias,$name
.

format DNS_REV =
@<<<<<<<<<<<<<<<<<<<<<<<<<<<<<   IN   PTR   @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$revip,$name
.
}

################################################################################
#
# Connect to the Database
#
################################################################################
sub connectdb
{
   local($dbname) = @_;

   if (!($dbh = Mysql->Connect($host, $dbname, $user, $pass,)))
   {
      die "[connectdb] Database connect failed: $Mysql::db_errstr\n";
   }
}

################################################################################
#
# Determine what to do based upon the environmental variables.
#
################################################################################
sub main
{
   &web_get();
   if ($FORM{'changelog'})
   {
      &set_up_page("CHANGE LOG for Version $VERSION");
      &show_changelog();
      &end_page();
      exit;
   }
   if ($FORM{'logout'})
   {
      &logout();
      exit;
   }
   &db_settings();
   &connectdb("@@DB@@");
   $USER = &get_user();
   $ACCESS = &get_access($USER);
   if ($FORM{'help'})
   {
      &set_up_page("Proper Syntax and Limitations for $FORM{'describe'}");
      &help($FORM{'describe'});
      print "<BR><CENTER><FONT SIZE=-1><A HREF=\"javascript:close_window()\">Click To Close This Window</A></CENTER>\n";
      &end_page();
      exit;
   }
   &defaults();
   &format_styles();
   @fields=&get_db_fields();
   if ($FORM{"function"} eq "Query")
   {
      @entry = &search_db($FORM{"Query"}, @fields);
      if (!@entry) { &show_options(1, @fields); }
      &set_up_page("Query Results for '$FORM{'Query'}':");
      &display("query", @fields);
      &end_page();
      exit;
   }
   if ($FORM{"function"} eq "Modify")
   {
      @entry = &search_db($FORM{"Query"}, @fields);
      if (!@entry) { &show_options(1, @fields); }
      &modify(@fields);
   }
   if ($FORM{"function"} eq "Add")
   {
      &new_entry(@fields);
   }
   if ($FORM{"function"} eq "Delete")
   {
      @entry = &search_db($FORM{"Query"}, @fields);
      if (!@entry) { &show_options(1, @fields); }
      &remove_entry(@fields);
   }
   if ($FORM{"Print"})
   {
      &dump_database(@fields);
   }
   if ($FORM{"function"} eq "Move")
   {
      &move_ip();  
      $notice = 2;
   }
   if ($FORM{"submit"} eq "modify")
   {
      $FORM{'Machine_Name'} = &clean_entry($FORM{'Machine_Name'},0);
      $FORM{'Alias'} = &clean_entry($FORM{'Alias'},0);
      $FORM{'IP_Address'} = &clean_entry($FORM{'IP_Address'},1);
      $notice=20 if (&exists($FORM{'Machine_Name'},"Machine_Name"));
      $notice=21 if ( &exists($FORM{'IP_Address'}, "IP_Address") && !$notice);
      if ($FORM{'Hardware_Address'})
      {
         $notice=22 if (&exists($FORM{'Hardware_Address'}, "Hardware_Address") && !$notice);
      }
      $notice=23 if ( (!$FORM{'Machine_Name'}) || (!$FORM{'IP_Address'}));
      if (!$notice)
      {
         &modify_entry();
         $notice=2;
      }
   }
   if ($FORM{"submit"} eq "delete")
   {
      &delete_entry();
      $notice=2;
   }
   if ($FORM{"submit"} eq "add")
   {
      $FORM{'Machine_Name'} = &clean_entry($FORM{'Machine_Name'},0);
      $FORM{'Alias'} = &clean_entry($FORM{'Alias'},0);
      $FORM{'IP_Address'} = &clean_entry($FORM{'IP_Address'},1);
      $notice=20 if (&exists($FORM{'Machine_Name'},"Machine_Name"));
      $notice=21 if ( &exists($FORM{'IP_Address'}, "IP_Address") && !$notice);
      if ($FORM{'Hardware_Address'})
      {
         $notice=22 if (&exists($FORM{'Hardware_Address'}, "Hardware_Address") && !$notice);
      }
      $notice=23 if ( (!$FORM{'Machine_Name'}) || (!$FORM{'IP_Address'}));
      if (!$notice)
      {
         &add_entry();
         $notice=2;
      }
   }
   if ($FORM{'PushMaps'})
   {
      &update_maps();
      &push_maps();
      $notice=3;
   }
   if ($FORM{'Reports'})
   {
      &subnet_usage();
   }
   if ($FORM{'Admin'})
   {
      &administer();
      exit;
   }
   if ($FORM{'UPDATE_CONF'})
   {
      if ( $FORM{'Update'} eq "Cancel")
      {
         $notice=5;
      }
      else
      {
         &update_config();
         $notice=4;
      }
   }
   if ($FORM{'User_Mod'})
   {
      &modify_user();
      $notice=6;
   }
   &show_options($notice,@fields);
}
 
################################################################################
#
# Print the top of the page, starting things off.
#
################################################################################
sub set_up_page()
{
   ($title) = @_;

   print <<END
Content-type: text/html


<HTML>
<HEAD>
   <TITLE>DNS Machine Database</TITLE>
   <SCRIPT LANGUAGE="Javascript">
   function openDescription(placement, height)
   {
      if (!height)
         height = 240;
            
       window.name='cd0ContentFrame';
END
;
   print "       var desString = '$my_name?help=1&describe=' + placement;\n";
   print <<END
       var desWin = window.open(desString,'description', 'scrollbars=yes,location=no,status=no,menubar=no,height='+height+',width=420,resizable=yes');
       desWin.focus();
    }

   function close_window()
   {
      aWindow=window.close();
   }
</SCRIPT>
</HEAD>
<BODY TOPMARGIN="0" LEFTMARGIN="0" MARGINWIDTH="0" MARGINHEIGHT="0" BGCOLOR=#ffffff>
END
;
   if (!$title)
   {
      print "<CENTER><IMG SRC=\"@@IMAGES@@/inventory.jpg\"></CENTER>";
   }
   else
   {
      print "<H1>$title</H1>";
   }
   print "<HR SIZE=1 NOSHADE>\n";
}

################################################################################
#
# Print the bottom of the page, closing things out.
#
################################################################################
sub end_page()
{
   local($show) = @_;

   print <<END
<HR SIZE=1 NOSHADE>
<TABLE WIDTH=100% CELLSPACING=0 CELLPADDING=0 BORDER=0>
<TR>
   <TD ALIGN=LEFT VALIGN=TOP WIDTH=45%>
<FONT SIZE=1>
&nbsp;&nbsp; This page, and all contents, &copy; 1999-2001 by <A HREF="mailto:rwest\@wesmo.com">Rich West</A>
</FONT>
   </TD>
END
;
   if ($show)
   {
      print <<END
   <TD VALIGN=TOP ALIGN=CENTER>
      <FONT SIZE=1><A HREF="$my_name?logout=1">Log Out</A>
   </FONT>
   </TD>
END
;
   }
   print <<END
   <TD ALIGN=RIGHT VALIGN=TOP WIDTH=45%>
<FONT SIZE=1>
<P ALIGN=RIGHT>Version: <A HREF="$my_name?changelog=1">$VERSION</A>
</FONT>
   </TD>
</TR>
</TABLE>
</BODY>
</HTML>
END
;
}

################################################################################
#
# Display a login screen, asking for a name and password.  It is rudimentary,
# but it works.
#
################################################################################
sub get_user()
{
   local ($USER, %cookies, $sth, $qs, $stored_password, $new, $seed, $PASSWORD);

   %cookies = fetch CGI::Cookie;
   if ( $cookies{'MachineDatabase'} )
   {
      $USER = $cookies{'MachineDatabase'}->value;
      $set_cookie = new CGI::Cookie(-name => 'MachineDatabase',
          -value => $USER,
          -expires => '+8h');
      print "Set-Cookie: $set_cookie\n";
      return ($USER);
   }
   if ($FORM{'user'})
   {
      $qs = "SELECT Password FROM users WHERE Username='$FORM{'user'}'";
      $sth = $dbh->Query($qs)
         or &notification("[get_user] Query had some problem: $Mysql::db_errstr\n$qs\n");
      @row=$sth->FetchRow;
      $stored_password = $row[0];
      if ($stored_password)
      {
         $seed = (split(//, $stored_password))[0] . (split(//, $stored_password))[1];
         $new = crypt($FORM{'password'}, $seed);
         if ($stored_password eq $new)
         {
            $USER = $FORM{'user'};
            $PASSWORD = $FORM{'password'};
            $set_cookie = new CGI::Cookie(-name => 'MachineDatabase',
                -value => $USER,
                -expires => '+8h');
            print "Set-Cookie: $set_cookie\n";
            return ($USER);
         }
      }
   }
   &set_up_page("<CENTER>You MUST Login in order to Access this Database.</CENTER>");
   print <<END
<FORM METHOD="POST" ACTION="@@CGIBIN@@/$my_name">
<CENTER>
<TABLE BGCOLOR="#3a1919" CELLPADDING=0 CELLSPACING=0 BORDER=0>
<TR>
<TD COLSPAN=3><IMG SRC="@@IMAGES@@/top.gif" BORDER=0></TD>
</TR>
<TR>
<TD><IMG SRC="@@IMAGES@@/login.gif" BORDER=0></TD>
<TD><INPUT TYPE="text" NAME="user" SIZE=10></TD>
<TD>&nbsp;</TD>
</TR>
<TR>
<TD>
<IMG SRC="@@IMAGES@@/password.gif" BORDER=0>
</TD>
<TD><INPUT TYPE="password" NAME="password" SIZE=10></TD>
<TD>&nbsp;</TD>
</TR>
<TR>
<TD ALIGN=CENTER COLSPAN=3><INPUT TYPE="image" BORDER=0 SRC="@@IMAGES@@/log_in.gif" ALT="Log In!"></TD>
</TR>
<TR>
<TD COLSPAN=3><IMG SRC="@@IMAGES@@/bottom.gif" BORDER=0></TD>
</TR>
</TABLE>
</FORM>
</CENTER>
END
;
   &end_page(0);
   exit;
}

################################################################################
#
# Grab the user's access level
#
################################################################################
sub get_access
{
   local($USER, $qs, $sth, @row) = @_;

   $qs = "SELECT Access FROM users WHERE Username='$USER'";
   $sth = $dbh->Query($qs)
      or &notification("[get_access] Query had some problem: $Mysql::db_errstr\n$qs\n");
   @row=$sth->FetchRow;
   return ($row[0]);
}

################################################################################
#
# Blow away the user's cookie then bring them back to the login screen.
#
################################################################################
sub logout
{
   %cookies = fetch CGI::Cookie;
   if (exists $cookies{'MachineDatabase'})
   {
      my $pass_cookie;
      $pass_cookie = $cookies{'MachineDatabase'};
      $pass_cookie->expires('now');
      $pass_cookie->value('');
      print "Set-Cookie: ", $pass_cookie, "\n";
   }
   print "Content-type: text/html\n\n";
   print " <HTML> <HEAD> <META HTTP-EQUIV=\"REFRESH\" CONTENT=\"2 ; URL=$my_name\">\n</HEAD>";
   print "<CENTER><H2><FONT COLOR=RED>Processing Log Out...</FONT></H2></CENTER>\n";
   print "</BODY>";
}
################################################################################
#
# A rather standard routine to grab the environmental variables and drop them
# into a variable form of $FORM{'name'}.
#
################################################################################
sub web_get()
{
   ######cut here ###################################################
   # Get the input
   if ($ENV{'CONTENT_LENGTH'})
   {
      read(STDIN, $buffer, $ENV{'CONTENT_LENGTH'});
   }
   # Split the name-value pairs
   if ($buffer) {
       @pairs = split(/&/, $buffer);
   }
   else {
 
       @pairs = split(/&/,$ENV{'QUERY_STRING'});
   }
 
   foreach $pair (@pairs)
   {
       ($name, $value) = split(/=/, $pair);
       # Un-Webify plus signs and %-encoding
       $value =~ tr/+/ /;
       $value =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;
       $value =~ s/&lt/</g;
       $value =~ s/&gt/>/g;
       # Stop people from using subshells to execute commands
       # Not a big deal when using sendmail, but very important
       # when using UCB mail (aka mailx).
       $value =~ s/~!/ ~!/g;
       # Uncomment for debugging purposes
       #print "A gnarled troll tells you that  was set to $value<br>";
       $FORM{$name} = $value;
   }
#### ----cut here
}

################################################################################
#
# In a generic and dynamic format, show a pull-down selection of all of the
# available fields.
#
################################################################################
sub show_fields
{
   local($type,@fields) = @_;

   print "<SELECT NAME=\"$type\">\n";
   foreach $count (0..$#fields)
   {
      next if ($noshow{$fields[$count]});
      $_ = $fields[$count];
      s/_/ /;
      print "<OPTION";
      if ($FORM{$type} eq $fields[$count]) { print " SELECTED"; }
      print " VALUE=\"$fields[$count]\"> $_\n";
   }
   print "</SELECT>\n";
}

################################################################################
#
# Generate the initial welcome page, which shows all of the possible options
# and searches that can be done on the master DNS database.
#
################################################################################
sub show_options
{
   local($notice, @fields)=@_;

#   &set_up_page("DNS Machine Database Options"); # now a graphic
   &set_up_page("");
   &notification($notice) if ($notice);
   print <<END
<CENTER>
<TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0 WIDTH=80%>
<TR>
   <TD COLSPAN=3>
<FORM METHOD="POST" ACTION="@@CGIBIN@@/$my_name">
<SELECT NAME="function">
<OPTION VALUE="Query">Query
END
;
   if ($ACCESS < 2)
   {
      print <<END
<OPTION VALUE="Modify">Modify
<OPTION VALUE="Delete">Delete
END
;
   }
   print <<END
</SELECT>
 the Database for the Machine with the
END
;
   &show_fields("query_type",@fields);
print <<END
:
<INPUT TYPE="text" NAME="Query" value="$FORM{'Query'}" SIZE=8>
<INPUT TYPE="image" BORDER=0 SRC="@@IMAGES@@/go.gif" ALT="Go!">
</FORM>
   </TD>
</TR>
END
;
   %subs=&get_subnets();
   foreach $SUB (sort numerically keys %subs)
   {
      $subnets[++$#subnets] = "$SUB";
   }
   if ($ACCESS < 2)
   {
      print <<END
<TR>
   <TD COLSPAN=3>
<FORM METHOD="POST" ACTION="@@CGIBIN@@/$my_name">
<INPUT TYPE="submit" name="function" VALUE="Add">
a new entry to the database in the:
END
;
   &show_fields("subnet",@subnets);
print <<END
subnet.
<FONT SIZE=-2><INPUT TYPE="checkbox" name="specific">with a Specific IP</FONT>
</FORM>
   </TD>
</TR>
END
;
   }
   print <<END
<TR>
   <TD COLSPAN=3>
<FORM METHOD="POST" ACTION="@@CGIBIN@@/$my_name">
<INPUT TYPE="submit" name="Print" VALUE="Display">
 the 
END
;
   $subnets[++$#subnets] = "Entire";
   &show_fields("Range", @subnets);
   print " Database, sorting by ";
   &show_fields("query_type",@fields);
print <<END
</FORM>
   </TD>
</TR>
<TR>
   <TD ALIGN=LEFT>
<FORM METHOD="POST" ACTION="@@CGIBIN@@/$my_name">
Show the Subnet Usage
<INPUT TYPE="submit" name="Reports" VALUE="Reports">
</FORM>
   </TD>
   <TD ALIGN=CENTER>
END
;
   if ($ACCESS < 2)
   {
      print <<END
<FORM METHOD="POST" ACTION="@@CGIBIN@@/$my_name">
<INPUT TYPE="submit" name="PushMaps" VALUE="Push">
Maps
</FORM>
END
;
   }
   print <<END
   </TD>
   <TD ALIGN=RIGHT>
<FORM METHOD="POST" ACTION="@@CGIBIN@@/$my_name">
<INPUT TYPE="submit" name="Admin" VALUE="Administer"> Configuration/Settings
Maps
</FORM>
   </TD>
</TR>
</TABLE>
</CENTER>
END
;
   &end_page(1);
   exit;
}

################################################################################
#
# Search through the entire database and find the item specified.  It must
# match exactly.
#
################################################################################
sub search_db
{
   local($entry, @fields) = @_;

   if ($FORM{'query_type'} eq "Alias")
   {
      $id = &get_alias_id($entry); 
   }
   else
   {
      $id = &get_id($entry);
   }
   if ($id)
   {
      @machine_info = &get_ip($id);
      @alias[0] = &get_aliases($id);
      @owner = &get_owner($id);
      @specs = &get_specs($id);
      @machine_info = (@machine_info, @alias, @specs, @owner);
      return (@machine_info);
   }
   return(0);
}

################################################################################
#
# Get the id from the aliases database.
#
################################################################################
sub get_specs
{
   local($id) = @_;
   local($qs, $sth, @row);

   $qs = "SELECT OS,Serial_Number,Processor,Processor_Speed,Memory FROM specs WHERE id='$id'";
   $sth = $dbh->Query($qs)
      or &notification("[get_specs] Query had some problem: $Mysql::db_errstr\n$qs\n");
   @row=$sth->FetchRow;
   return (@row);
}

################################################################################
#
# Get the id from the aliases database.
#
################################################################################
sub get_alias_id
{
   local($entry) = @_;
   local($id);

   $qs = "SELECT id FROM aliases WHERE $FORM{'query_type'}='$entry'";
   $sth = $dbh->Query($qs)
      or &notification("[get_alias_id] Query had some problem: $Mysql::db_errstr\n$qs\n");
   @row=$sth->FetchRow;
   return ($row[0]);
}

################################################################################
#
# Get the database id for this entry.
#
################################################################################
sub get_id
{
   local($entry) = @_;
   local($qs);

   if ( ($FORM{'submit'} eq "add") || ($FORM{'submit'} eq "modify") )
   {
      $qs="SELECT id FROM name WHERE $FORM{'query_type'}='$entry'";
   }
   else
   {
      $qs="select name.id FROM name,owner WHERE $FORM{query_type}='$entry' AND name.id=owner.id"; 
   }
   $sth = $dbh->Query($qs)
      or &notification("[get_id] Query had some problem: $Mysql::db_errstr\n$qs\n");
   @row=$sth->FetchRow;
   return ($row[0]);
}

################################################################################
#
# Get the IP and other machine specific information
#
################################################################################
sub get_ip
{
   local($id) = @_;
   local($qs, $sth, @row);

   $qs="SELECT IP_Address,Machine_Name,Hardware_Address FROM name WHERE id=$id";
   $sth = $dbh->Query($qs)
      or &notification("[get_ip] Query had some problem: $Mysql::db_errstr\n$qs\n");
   @row=$sth->FetchRow;
   return (@row);
}

################################################################################
#
# Get all of the owner information.
#
################################################################################
sub get_owner
{
   local($id) = @_;
   local($qs, $sth, @row);

   $qs="SELECT Owner_Name,Location,Comment FROM owner WHERE id=$id";
   $sth = $dbh->Query($qs)
      or &notification("[get_owner] Query had some problem: $Mysql::db_errstr\n$qs\n");
   @row=$sth->FetchRow;
   return (@row);
}

################################################################################
#
# Get all of the aliases for this entry.
#
################################################################################
sub get_aliases
{
   local($id) = @_;
   local($qs, $sth, $aliases, @row);
                                    
   $qs="SELECT Alias FROM aliases WHERE id=$id";
   $sth = $dbh->Query($qs)
      or &notification("[get_aliases] Query had some problem: $Mysql::db_errstr\n$qs\n");
   @row=$sth->FetchRow;
   $aliases = $row[0];
   while (@row=$sth->FetchRow)
   {
      $aliases .= ",$row[0]";
   }
   return ("$aliases");
}

################################################################################
#
# Display the last time that the record was changed and what userid changed it.
#
################################################################################
sub show_last_modification
{
   local($id) = @_;
   local($qs, $sth);

   $qs = "SELECT who,date FROM changes WHERE id='$id'";
   $sth = $dbh->Query($qs)
      or &notification("[show_last_modification] Query had some problem: $Mysql::db_errstr\n$qs\n");
   @row=$sth->FetchRow;
   $day = substr($row[1], 6, 2);
   $month = substr($row[1], 4, 2);
   $year = substr($row[1], 0, 4);
   print "<CENTER><FONT SIZE=-1>This record was last updated on <B>$month/$day/$year</B> by <B>$row[0]</B>.</CENTER><BR>\n";
}

################################################################################
#
# A semi-generic routine which sets up the display for modifying, deleting,
# or adding an entry.
#
################################################################################
sub display
{
   local($type, @fields) = @_;
   local($count,@row,$sth,$qs);

   $count=0;
   &show_last_modification($id) if ($FORM{'function'} ne "Add");
   print "<TABLE CELLSPACING=5 CELLPADDING=10 BORDER=0>\n";
   print "<TR>\n";
   print "   <TD VALIGN=\"TOP\">\n";
   print "      <TABLE CELLSPACING=0 CELLPADDING=0 BORDER=0>\n";
   print "      <TH BGCOLOR=\"#aeaeae\" COLSPAN=3> Network Information </TH>\n";
   for ($count=0; $count<=3; $count++)
   {
      if (($fields[$count] eq "IP_Address") && ($FORM{'function'} eq "Add") && (!$FORM{'specific'}))
      {
         next;
      }
      print "      <TR>\n";
      print "         <TD>\n";
      $_ = $fields[$count];
      s/_/ /g;
      print "            $_ :\n";
      print "         </TD>\n";
      print "         <TD>\n";
      if (($type eq "modify") || ($type eq "new"))
      {
         print qq#   <INPUT TYPE="text" NAME="$fields[$count]" VALUE="$entry[$count]" SIZE=16>#;
      }
      else
      {
         print "$entry[$count] \n";
         print qq#   <INPUT TYPE="hidden" NAME="$fields[$count]" VALUE="$entry[$count]">\n# if ($type eq "delete");
      }
      print "         </TD>\n";
      print "         <TD>\n";
      if (($type eq "modify") || ($type eq "new"))
      {
         print qq#         <A HREF="javascript://" onClick="openDescription('$fields[$count]');"><img src="@@IMAGES@@/descriptionButton.gif" width="10" height="10" border="0" alt="Click here for more information about the $fields[$count] field"></a>#;
      }
      print "         </TD>\n";
      print "      </TR>\n";
   }
   print "      </TABLE>\n";
   print "   </TD>\n";
   print "   <TD VALIGN=\"TOP\">\n";
   print "      <TABLE CELLSPACING=0 CELLPADDING=0 BORDER=0>\n";
   print "      <TH BGCOLOR=\"#aeaeae\" COLSPAN=3> Machine Information </TH>\n";
   for ($count=4; $count<=8; $count++)
   {
      print "      <TR>\n";
      print "         <TD>\n";
      print "            $fields[$count] :\n";
      print "         </TD>\n";
      print "         <TD>\n";
      if (($type eq "modify") || ($type eq "new"))
      {
         if ($fields[$count] eq "OS")
         {
            $qs = "SELECT * FROM OS";
            $sth = $dbh->Query($qs)
               or &notification("[display] Query had some problem: $Mysql::db_errstr\n$qs\n");
            print "<SELECT NAME=\"OS\">\n";
            print "<OPTION VALUE=\"0\">None\n";
            while(@row=$sth->FetchRow)
            {
               print "<OPTION VALUE=\"$row[0]\"";
               print "SELECTED" if ($entry[$count] eq $row[0]);
               print ">$row[1]\n";
            }
            print "</SELECT>\n";
         }
         else
         {
            if ($fields[$count] eq "Processor")
            {
               $qs = "SELECT * FROM processor_types";
               $sth = $dbh->Query($qs)
                  or &notification("[display] Query had some problem: $Mysql::db_errstr\n$qs\n");
               print "<SELECT NAME=\"Processor\">\n";
               print "<OPTION VALUE=\"0\">None\n";
               while(@row=$sth->FetchRow)
               {
                  print "<OPTION VALUE=\"$row[0]\"";
                  print "SELECTED" if ($entry[$count] eq $row[0]);
                  print ">$row[1]\n";
               }
               print "</SELECT>\n";
            }
            else
            {
               print qq#   <INPUT TYPE="text" NAME="$fields[$count]" VALUE="$entry[$count]" SIZE=16>#;
            }
         }
      }
      else
      {
         if ($fields[$count] eq "OS")
         {
            $qs = "SELECT name FROM OS WHERE id='$entry[$count]'";
            $sth = $dbh->Query($qs)
               or &notification("[display] Query had some problem: $Mysql::db_errstr\n$qs\n");
            @row=$sth->FetchRow;
            print "$row[0] \n";
         }
         else
         {
            if ($fields[$count] eq "Processor")
            {
               $qs = "SELECT name FROM processor_types WHERE id='$entry[$count]'";
               $sth = $dbh->Query($qs)
                  or &notification("[display] Query had some problem: $Mysql::db_errstr\n$qs\n");
               @row=$sth->FetchRow;
               print "$row[0] \n";
            }
            else
            {
               print "$entry[$count] \n";
            }
         }
      }
      print "         </TD>\n";
      print "         <TD>\n";
      if (($type eq "modify") || ($type eq "new"))
      {
         print qq#         <A HREF="javascript://" onClick="openDescription('$fields[$count]');"><img src="@@IMAGES@@/descriptionButton.gif" width="10" height="10" border="0" alt="Click here for more information about the $fields[$count] field"></a>#;
      }
      print "         </TD>\n";
      print "      </TR>\n";
   }
   print "      </TABLE>\n";
   print "   </TD>\n";
   print "</TR>\n";
   print "<TR>\n";
   print "   <TD VALIGN=\"TOP\">\n";
   print "      <TABLE CELLSPACING=0 CELLPADDING=0 BORDER=0>\n";
   print "      <TH BGCOLOR=\"#aeaeae\" COLSPAN=3> Owner Information </TH>\n";
   for ($count=9; $count<=11; $count++)
   {
      print "      <TR>\n";
      print "         <TD>\n";
      print "            $fields[$count] :\n";
      print "         </TD>\n";
      print "         <TD>\n";
      if (($type eq "modify") || ($type eq "new"))
      {
         print qq#   <INPUT TYPE="text" NAME="$fields[$count]" VALUE="$entry[$count]" SIZE=16>#;
      }
      else
      {
         print "$entry[$count] \n";
      }
      print "         </TD>\n";
      print "         <TD>\n";
      if (($type eq "modify") || ($type eq "new"))
      {
         print qq#         <A HREF="javascript://" onClick="openDescription('$fields[$count]');"><img src="@@IMAGES@@/descriptionButton.gif" width="10" height="10" border="0" alt="Click here for more information about the $fields[$count] field"></a>#;
      }
      print "         </TD>\n";
      print "      </TR>\n";
   }
   print "      </TABLE>\n";
   print "   </TD>\n";
   print "</TR>\n";
   print "</TABLE>\n";
   if (! (($type eq "modify") || ($type eq "new") || ($type eq "delete")) )
   {
      print <<EOF
<BR>
<FORM METHOD="POST" ACTION="@@CGIBIN@@/$my_name">
<TABLE CELLPADDING=0 CELLSPACING=10 BORDER=0>
<TR>
   <TD>
      <INPUT TYPE=SUBMIT NAME="function" VALUE="Modify">
   </TD>
   <TD>
      <INPUT TYPE=SUBMIT NAME="function" VALUE="Delete">
   </TD>
   <TD>
      <INPUT TYPE=SUBMIT NAME="function" VALUE="Move"> to the 
EOF
;
   %subs=&get_subnets();
   foreach $SUB (sort numerically keys %subs)
   {
      $subnets[++$#subnets] = "$SUB";
   }
   &show_fields("subnet",@subnets);
print <<EOF
subnet.
   </TD>
</TR>
</TABLE>
<INPUT TYPE=HIDDEN NAME="query_type" VALUE="$FORM{'query_type'}">
<INPUT TYPE=HIDDEN NAME="Query" VALUE="$FORM{'Query'}">
</FORM>
EOF
;
   }
}
 
################################################################################
#
# Get all of the valid fields in the database.  This is grabbed from the first
# line in the database, and is entirely dynamic.
#
################################################################################
sub get_db_fields
{
   local($field,$count);
   local($output,$i,$j,@row2,$table,@tables,@db);

   $i=0;
   @tables = ("name", "aliases", "specs", "owner");
   for ($j=0; $j<=$#tables; $j++)
   {
      $qs="SHOW FIELDS FROM $tables[$j]";
      $sth = $dbh->Query($qs)
      or &notification("[get_db_fields] Query had some problem: $Mysql::db_errstr\n$qs\n");
      while(@row=$sth->FetchRow)
      {
         next if ($row[0] eq "id");
         $field[$i] = $row[0];
         $db[$i] = $table;
         $i++;
      }
   }
   return (@field);
}

################################################################################
#
# Generate the page to get the information to modify an entry.
#
################################################################################
sub modify
{
   local(@fields) = @_;

   &set_up_page("Modify '$FORM{'Query'}':");
   print <<END
<FORM METHOD="POST" ACTION="@@CGIBIN@@/$my_name">
<INPUT TYPE="hidden" name="submit" value="modify">
<INPUT TYPE="hidden" name="id" value=\"$id\">
END
;
   &display("modify", @fields);
   print <<END
<INPUT TYPE="submit" VALUE="Commit Modification">
</FORM>
END
;
   &end_page(1);
   exit;
}

################################################################################
#
# Generate the page to get the information to add an entry.
#
################################################################################
sub new_entry
{
   local(@fields)=@_;
   local($ip);

   &set_up_page("New Entry Form");
   print <<END
<FORM METHOD="POST" ACTION="@@CGIBIN@@/$my_name">
<INPUT TYPE="hidden" name="submit" value="add">
<INPUT TYPE="hidden" name="query_type" value="IP_Address">
END
;
   print qq#<INPUT TYPE="hidden" name="specific" value="true">\n# if ($FORM{'specific'});
   &display("modify", @fields);
   $ip=&new_ip($FORM{subnet}) if (!$FORM{'specific'});
   print qq#<INPUT TYPE="hidden" name="IP_Address" value="$ip">\n# if (!$FORM{'specific'});
   print qq#<INPUT TYPE="submit" VALUE="Commit Addition">\n</FORM>#;
   &end_page(1);
   exit;
}

################################################################################
#
# Strip out some of the more popular invalid characters.
#
################################################################################
sub clean_entry
{
   local ($variable,$flag) = @_;

   $_ = $variable;
   if ($flag)
   {
      s/\ //g;
   }
   else
   {
      s/\ /-/g;
      s/\_/-/g;
      s/\*/-/g;
      s/\&/-/g;
      s/\%/-/g;
      s/\$/-/g;
      s/\#/-/g;
      s/\@/-/g;
      s/\!/-/g;
      s/\`/-/g;
      s/\~/-/g;
      s/\'/-/g;
      s/\"/-/g;
      s/\:/-/g;
      s/\;/-/g;
   }
   return($_);
}

################################################################################
#
# Get the id of the entry based upon the initial query, set the proper
# $FORM variables for future use, then update the record to reflect the
# new IP address.
#
################################################################################
sub move_ip
{
   local ($id,$qs,$sth,@row);

   $id = &get_id($FORM{'Query'});
   $FORM{'IP_Address'} = &new_ip($FORM{'subnet'});
   $qs = "SELECT Machine_Name FROM name WHERE id='$id'";
   $sth = $dbh->Query($qs)
      or &notification("[move_ip] Query had some problem: $Mysql::db_errstr\n$qs\n");
   @row=$sth->Fetchrow;
   $FORM{'Machine_Name'} = $row[0];
   $qs = "UPDATE name SET IP_Address='$FORM{'IP_Address'}' WHERE id='$id'";
   $sth = $dbh->Query($qs)
      or &notification("[move_ip] Query had some problem: $Mysql::db_errstr\n$qs\n");
   &update_changelog($id);
}

################################################################################
#
# Update all references to the entry according to what was entered in the
# form.  Actually, we update everything except the aliases.  With the aliases,
# it is easier to remove them all and re-add them.
#
################################################################################
sub modify_entry
{
   local($sth, $qs, $date);

   $qs = "UPDATE name SET IP_Address='$FORM{'IP_Address'}',Machine_Name='$FORM{'Machine_Name'}',Hardware_Address='$FORM{'Hardware_Address'}' WHERE id='$FORM{'id'}'";
   $sth = $dbh->Query($qs)
      or &notification("[modify_entry] Query had some problem: $Mysql::db_errstr\n$qs\n");
   $qs = "UPDATE owner SET Owner_Name='$FORM{'Owner_Name'}',Location='$FORM{'Location'}',Comment='$FORM{'Comment'}' WHERE id='$FORM{'id'}'";
   $sth = $dbh->Query($qs)
      or &notification("[modify_entry] Query had some problem: $Mysql::db_errstr\n$qs\n");
   $qs = "UPDATE specs SET OS='$FORM{'OS'}',Serial_Number='$FORM{'Serial_Number'}',Processor='$FORM{'Processor'}',Processor_Speed='$FORM{'Processor_Speed'}' WHERE id='$FORM{'id'}'";
   $sth = $dbh->Query($qs)
      or &notification("[modify_entry] Query had some problem: $Mysql::db_errstr\n$qs\n");
   $qs="DELETE FROM aliases WHERE id='$FORM{'id'}'";
   $sth = $dbh->Query($qs)
      or &notification("[modify_entry] Query had some problem: $Mysql::db_errstr\n$qs\n");
   if ($FORM{'Alias'})
   {
      @aliases = split(/,/, $FORM{'Alias'});
      for($i=0; $i<=($#aliases); $i++)
      {
         $qs = "INSERT INTO aliases VALUES ('$FORM{'id'}', '$aliases[$i]')";
         $sth = $dbh->Query($qs)
            or &notification("[modify_entry] Query had some problem: $Mysql::db_errstr\n$qs\n");
      }
   }
   &update_changelog($FORM{'id'});
}

################################################################################
#
# Record who made the last change and on what day.
#
################################################################################
sub update_changelog
{
   local($id) = @_;
   local($qs,$sth);

   $date = get_date();
   $qs = "UPDATE changes SET who='$USER',date='$date' WHERE id='$id'";
   $sth = $dbh->Query($qs)
      or &notification("[modify_entry] Query had some problem: $Mysql::db_errstr\n$qs\n");
}

################################################################################
#
# Physically add an entry into the database, defined by what was previously
# entered into the on-line form (the subnet).  As of this time, no error
# checking is done on the data entered.
#
################################################################################
sub add_entry
{
   local($sth, $qs, $date);

   $qs = "INSERT INTO name VALUES ('', '$FORM{'IP_Address'}', '$FORM{'Machine_Name'}', '$FORM{'Hardware_Address'}')";
   $sth = $dbh->Query($qs)
      or &notification("[add_entry] Query had some problem: $Mysql::db_errstr\n$qs\n");
   $FORM{query_type}="IP_Address";
   $id = &get_id($FORM{'IP_Address'});
   $qs = "INSERT INTO specs VALUES ('$id', '$FORM{'OS'}', '$FORM{'Serial_Number'}', '$FORM{'Processor'}', '$FORM{'Processor_Speed'}', '$FORM{'Memory'}')";
   $sth = $dbh->Query($qs)
      or &notification("[add_entry] Query had some problem: $Mysql::db_errstr\n$qs\n");
   if ($FORM{'Alias'})
   {
      @aliases = split(/,/, $FORM{'Alias'});
      for($i=0; $i<=($#aliases); $i++)
      {
         $qs = "INSERT INTO aliases VALUES ('$id', '$aliases[$i]')";
         $sth = $dbh->Query($qs)
            or &notification("[add_entry] Query had some problem: $Mysql::db_errstr\n$qs\n");
      }
   }
   $qs = "INSERT INTO owner VALUES ('$id', '$FORM{'Owner_Name'}', '$FORM{'Location'}', '$FORM{'Comment'}')";
   $sth = $dbh->Query($qs)
      or &notification("[add_entry] Query had some problem: $Mysql::db_errstr\n$qs\n");
   $date = &get_date();
   $qs = "INSERT INTO changes VALUES ('$id', '$USER', '$date')";
   $sth = $dbh->Query($qs)
      or &notification("[add_entry] Query had some problem: $Mysql::db_errstr\n$qs\n");
}

################################################################################
#
# Physically remove an entry from the database, defined by what was previously
# entered into the on-line form.  The item must match exactly.
#
################################################################################
sub delete_entry
{
   local($split_entry,$count,$qs);

   $qs="DELETE FROM name WHERE id='$FORM{'id'}'";
   $sth = $dbh->Query($qs)
      or &notification("[delete_entry] Query had some problem: $Mysql::db_errstr\n$qs\n");
   $qs="DELETE FROM aliases WHERE id='$FORM{'id'}'";
   $sth = $dbh->Query($qs)
      or &notification("[delete_entry] Query had some problem: $Mysql::db_errstr\n$qs\n");
   $qs="DELETE FROM owner WHERE id='$FORM{'id'}'";
   $sth = $dbh->Query($qs)
      or &notification("[delete_entry] Query had some problem: $Mysql::db_errstr\n$qs\n");
   $qs="DELETE FROM specs WHERE id='$FORM{'$id'}'";
   $sth = $dbh->Query($qs)
      or &notification("[delete_entry] Query had some problem: $Mysql::db_errstr\n$qs\n");
   $qs="DELETE FROM changes WHERE id='$FORM{'$id'}'";
   $sth = $dbh->Query($qs)
      or &notification("[delete_entry] Query had some problem: $Mysql::db_errstr\n$qs\n");
}

################################################################################
#
# Generate the delete page based upon the previously entered key, asking
# for an absoloute commital to the deletion.
#
################################################################################
sub remove_entry
{
   local(@fields)=@_;

   &set_up_page("Delete '$FORM{'Query'}':");
   print <<END
<FORM METHOD="POST" ACTION="@@CGIBIN@@/$my_name">
<INPUT TYPE="hidden" name="submit" value="delete">
<INPUT TYPE="hidden" name="id" value=\"$id\">
<INPUT TYPE="hidden" name="Machine_Name" value=\"$entry[2]\">
END
;
   &display("delete", @fields);
   print <<END
<INPUT TYPE="submit" VALUE="Commit Deletion">
</FORM>
END
;
   &end_page(1);
   exit;
}

################################################################################
#
# Display the entire contents of the database.
#
################################################################################
sub dump_database
{
   local($count, @fields, $bgcolor);

   &set_up_page("Listing of Database Contents");
   @fields=get_db_fields();
   print "<TABLE BORDER=2 CELLPADDING=0 CELLSPACING=0>\n<TR>\n";
   $bgcolor="#EEEEEE";
print "   <TD bgcolor=$bgcolor ALIGN=CENTER><B><FONT SIZE=-1>Delete</FONT></B></TD>\n";

   foreach $count (0..$#fields)
   {
      next if ($noshow{$fields[$count]});
      $location = $count if ($FORM{'query_type'} eq $fields[$count]);
      $_ = $fields[$count];
      s/_/ /;
      print <<EOF
   <TD bgcolor=$bgcolor>
      <A HREF="$my_name?Print=$FORM{'Print'}&Range=$FORM{'Range'}&query_type=$fields[$count]"><B>$_</B></A>
   </TD>
EOF
;
   }
   print "</TR>\n";
   if ($FORM{'Range'} eq "Entire")
   {
      %subnets=&get_subnets();
   }
   else
   {
      %subnets=("$FORM{'Range'}");
   }
   foreach $SUBNET (sort numerically keys %subnets)
   {
      if ($FORM{'query_type'} eq "IP_Address")
      {
         &dump_by_ip($SUBNET);
      }
      else
      {
         &dump_else($SUBNET);
      }
   }
   print "</TABLE>\n";
   &end_page(1);
   exit;
}

################################################################################
#
# Display the contents of the database, sorting by IP.  This is a separate
# routine because sorting by a period-separated numeric is a pain in the butt.
# What we do here is remove the dots, pad the octets, mash the number together,
# and sort it numerically based upon that huge number.
#
################################################################################
sub dump_by_ip
{
   local($SUBNET) = @_;
   local($qs,$qs2,$sth,$sth2,@row,@row2,$a,$b,$c,$d,%ip_list,$ip_addr,$qs);

   $qs = "SELECT IP_Address FROM name WHERE IP_Address LIKE '$SUBNET.%'";
   $sth = $dbh->Query($qs)
      or &notification("[dump_by_ip] Query had some problem: $Mysql::db_errstr\n$qs\n");
   while (@row=$sth->Fetchrow)
   {
      ($a,$b,$c,$d) = split(/\./,$row[0]);
      $temp = &padit($a) . &padit($b) . &padit($c) . &padit($d);
      $ip_list{$temp} = 1;
   }
   foreach $ip_addr (sort numerically keys %ip_list)
   {
   $qs = "SELECT DISTINCT name.id,name.IP_Address,name.Machine_Name,name.Hardware_Address,owner.id,owner.Owner_Name,owner.Location FROM name,specs,owner WHERE name.id=owner.id AND name.id=specs.id AND IP_Address LIKE '" . &unpadit($ip_addr) . "'";
      $sth = $dbh->Query($qs)
         or &notification("[dump_by_ip] Query had some problem: $Mysql::db_errstr\n$qs\n");
      @row=$sth->Fetchrow;
      $qs2 = "SELECT * FROM aliases WHERE id='$row[0]'";
      $sth2 = $dbh->Query($qs2)
         or &notification("[dump_by_ip] Query had some problem: $Mysql::db_errstr\n$qs2\n");
      if (@row2=$sth2->Fetchrow)
      {
         $row[4] = $row2[1];   
      }
      else
      {
          $row[4] = "&nbsp;";
      }
      while (@row2=$sth2->Fetchrow)
      {
         $row[4] .= ",<BR>$row2[1]";
      }
      if ($bgcolor eq "#EEEEEE")
      {
         $bgcolor="#FFFFFF";
      }
      else
      {
         $bgcolor="#EEEEEE";
      }
      print "<TR>\n";
      for ($count=0; $count<=$#row; $count++)
      {
         print "   <TD bgcolor=$bgcolor>\n";
         if ($count eq "1")
         {
            print qq#      <A HREF="@@CGIBIN@@/$my_name?query_type=IP_Address&function=Modify&Query=$row[$count]">$row[$count]</A>\n# if ($count == 1);
         }
         else
         {
            if ($count eq "0")
            {
               print qq#      <CENTER><A HREF="@@CGIBIN@@/$my_name?query_type=IP_Address&function=Delete&Query=$row[1]"><IMG SRC="@@IMAGES@@/delete.gif" BORDER=0></A></CENTER>\n#;
            }
            else
            {
               print "$row[$count]&nbsp;\n";
            }
         }
         print "   </TD>\n";
      }
      print "</TR>\n";
   }
}

################################################################################
#
# Display the database sorting by the field we have selected.
#
################################################################################
sub dump_else
{
   local($SUBNET) = @_;
   local($qs,$qs2,$sth,$sth2,@row,@row2);

   $qs = "SELECT DISTINCT name.id,name.IP_Address,name.Machine_Name,name.Hardware_Address,owner.id,owner.Owner_Name,owner.Location FROM name,specs,owner WHERE name.id=owner.id AND name.id=specs.id AND IP_Address LIKE '$SUBNET.%' ORDER BY '$FORM{'query_type'}'";
   $sth = $dbh->Query($qs)
      or &notification("[dump_else] Query had some problem: $Mysql::db_errstr\n$qs\n");
   while (@row=$sth->Fetchrow)
   {
#      $qs = "SELECT name FROM OS WHERE id='$row[4]'";
#      $sth2 = $dbh->Query($qs)
#         or &notification("[dump_else] Query had some problem: $Mysql::db_errstr\n$qs\n");
#      @row2=$sth2->Fetchrow;
#      $row[4] = $row2[0];
      $qs2 = "SELECT * FROM aliases WHERE id='$row[0]'";
      $sth2 = $dbh->Query($qs2)
         or &notification("[dump_else] Query had some problem: $Mysql::db_errstr\n$qs2\n");
      if (@row2=$sth2->Fetchrow)
      {
         $row[4] = $row2[1];   
      }
      else
      {
          $row[4] = "&nbsp;";
      }
      while (@row2=$sth2->Fetchrow)
      {
         $row[4] .= ",<BR>$row2[1]";
      }
      if ($bgcolor eq "#EEEEEE")
      {
         $bgcolor="#FFFFFF";
      }
      else
      {
         $bgcolor="#EEEEEE";
      }
      print "<TR>\n";
      for ($count=0; $count<=$#row; $count++)
      {
         print "   <TD bgcolor=$bgcolor>\n";
         if ($count eq "0")
         {
            print qq#      <CENTER><A HREF="@@CGIBIN@@/$my_name?query_type=IP_Address&function=Delete&Query=$row[1]"><IMG SRC="@@IMAGES@@/delete.gif" BORDER=0></A></CENTER>\n#;
         }
         else
         {
            if ($count eq "1")
            {
               print qq#      <A HREF="@@CGIBIN@@/$my_name?query_type=IP_Address&function=Modify&Query=$row[$count]">$row[$count]</A>\n#;
            }
            else
            {
               print "$row[$count]&nbsp;\n";
            }
         }
         print "   </TD>\n";
      }
      print "</TR>\n";
   }
}
################################################################################
#
# Tkae the incoming number (an IP), break it apart, pad each with leading
# zeros, and put it together into one numeric string.
#
################################################################################
sub padit
{
   local($num) = @_;

   $num="00" . $num if ($num<10);
   $num="0" . $num if (($num>9) && ($num<100));
   return ($num);
}

################################################################################
#
# Take the incoming number (a padded IP), and break it back down into its
# full IP.
#
################################################################################
sub unpadit
{
   local($num) = @_;

   $a = (substr($num, 0, 3)) + 0;
   $b = (substr($num, 3, 3)) + 0;
   $c = (substr($num, 6, 3)) + 0;
   $d = (substr($num, 9, 3)) + 0;
   return ("$a.$b.$c.$d");
}
   

################################################################################
#
# Get the next available IP for that subnet.
#
################################################################################
sub new_ip
{
   local ($SUBNET) = @_;
   local ($max, $ip);

   $qs = "SELECT IP_Address FROM name WHERE IP_Address LIKE '$SUBNET.%'";
   $sth = $dbh->Query($qs)
      or &notification("[new_ip] Query had some problem: $Mysql::db_errstr\n$qs\n");
   while (@row=$sth->FetchRow)
   {
      ($a,$b,$c,$d) = split(/\./, $row[0]);
      $ip = &padit($a) . &padit($b) . &padit($c) . &padit($d);
      $ips{$ip} = 1 if ($ip != 0);
   }  
   $max=$RANGE[0];
   foreach $addr (sort numerically keys %ips)
   {  
      $ip=substr($addr, 9);
      next if (($ip < $RANGE[0]) || ($ip > $RANGE[1]));
      ($max++) if ($ip == $max);
   }
   if ($max > 254)
   {
      &error_out("The $subnet subnet is full.  Please select another subnet.");
   }
   return ("$SUBNET.$max");
}

################################################################################
#
# Parse the entire database, looking for all of the possible valid subnets,
# and dump them into an array.
#
################################################################################
sub get_subnets
{
   local($sub);

   $qs="SELECT IP_Address FROM name";
   $sth = $dbh->Query($qs)
      or &notification("[get_subnets] Query had some problem: $Mysql::db_errstr\n$qs\n");
   while (@row=$sth->FetchRow)
   {
      $sub = (split(/\./, $row[0]))[0] . "." . (split(/\./, $row[0]))[1] . "." . (split(/\./, $row[0]))[2];
      $subnets{$sub} = 1 if ($sub != 0);
   }
   return (%subnets);
}

################################################################################
#
# Generate the real DNS files from the master database.  Break out the
# information, in a fully subnet-sorted order, and put them in the hosts
# and reverse database file format.
#
################################################################################
sub update_maps()
{
   local ($first,$second,$third,$fourth,%subnets);

   %subnets=&get_subnets();
   &build_all_dns(%subnets) if ($NAMED_DIR);
   &build_yp_hosts(%subnets) if ($NIS_DIR);
   &build_yp_ethers(%subnets) if ($NIS_DIR);
   &build_dhcp() if ($DHCPD_DIR);
}

################################################################################
#
# Generate the dhcpd table.
#
################################################################################
sub build_dhcp
{
   local($qs);
   $qs = "SELECT dhcpd_header FROM conf WHERE domainname='$DOMAINNAME'"; 
   $sth = $dbh->Query($qs)
      or &notification("[build_dhcp] Query had some problem: $Mysql::db_errstr\n$qs\n");
   @row=$sth->FetchRow;
   $_ = $row[0];
   s/
//g;
   $header = $_;
   $qs = "SELECT IP_Address,Machine_Name,Hardware_Address FROM name WHERE Hardware_Address!='' AND Hardware_Address!='0'";
   $sth = $dbh->Query($qs)
      or &notification("[build_dhcp] Query had some problem: $Mysql::db_errstr\n$qs\n");
   while (@row=$sth->FetchRow)
   {
      $content .= "
host $row[1]
{
   option host-name \"$row[1]\";
   hardware ethernet $row[2];
   fixed-address $row[0];
}
";
   }
   open(DHCP, "> $TMP/dhcpd.conf") || die "Could not write dhcpd.conf\n";
   print DHCP $header;
   print DHCP $content;
   close (DHCP);
}

################################################################################
#
# Grab the header information for YP from the 'conf' database.
#
################################################################################
sub yp_header
{
   local($qs);
   $qs = "SELECT yp_header FROM conf WHERE domainname='$DOMAINNAME'"; 
   $sth = $dbh->Query($qs)
      or &notification("[yp_header] Query had some problem: $Mysql::db_errstr\n$qs\n");
   @row=$sth->FetchRow;
   $_ = $row[0];
   s/^M//g;
   print NEWYP $row[0];
}

################################################################################
#
# Build the hosts file, using the header from the database and separating
# each subnet to keep things clean and easy to visually read.
#
################################################################################
sub build_yp_hosts
{
   local(%subnets) = @_;

   open(NEWYP, "> $TMP/hosts") || die &error_out("Could not write temporary hosts file.");
   &yp_header();
   foreach $SUBNET (sort numerically keys %subnets)
   {
      print NEWYP "#------------------------------\n";
      print NEWYP "# Subnet $SUBNET\n";
      print NEWYP "#------------------------------\n";
      $qs = "SELECT * FROM name WHERE IP_Address LIKE '$SUBNET.%'";
      $sth = $dbh->Query($qs)
         or &notification("[build_yp_hosts] Query had some problem: $Mysql::db_errstr\n$qs\n");
      while (@row=$sth->FetchRow)
      {
         ($a,$b,$c,$d) = split(/\./, $row[1]);
         $ip = &padit($a) . &padit($b) . &padit($c) . &padit($d);
         $ip = $ip . ":" . $row[2] . ":" . $row[0];
         $ips{$ip} = 1 if ($ip != 0);
      }
      foreach $addr (sort numerically keys %ips)
      {
         ($ip, $name, $id) = split(/:/, $addr);
         $aliases = &get_aliases($id);
         $ip = unpadit($ip);
         print NEWYP "$ip	$name $name.$DOMAINNAME";
         foreach $alias (split(/\,/, $aliases))
         {
            print NEWYP " $alias";
         }
         print NEWYP "\n";
         delete $ips{$addr};
      }
   }
   close(NEWYP);
}

################################################################################
#
# Build the ethers file.  Grab everything because it is easier..
#
################################################################################
sub build_yp_ethers
{
   local($qs, @row);

   $qs = "SELECT Hardware_Address,Machine_Name FROM name";
   open(NEWYP, "> $TMP/ethers") || die &error_out("Could not write temporary hosts file.");
   $sth = $dbh->Query($qs)
      or &notification("[build_yp_ethers] Query had some problem: $Mysql::db_errstr\n$qs\n");
   while (@row=$sth->FetchRow)
   {
      print NEWYP "$row[0]	$row[1]\n" if ($row[0]);
   }
   close(NEWYP);
}

################################################################################
#
# Get the current date and put it in the format YYYYMMDD.
#
################################################################################
sub get_date
{
   local($date,$month,$day,$year);

   $year = (localtime(time()))[5] + 1900;
   $month = (localtime(time()))[4] + 1;
   if ($month < 10)
   {
      $month = "0$month";
   }
   $day = (localtime(time()))[3];
   if ($day < 10)
   {
      $day = "0$day";
   }
   $date = $year . $month . $day;
   return ($date);
}

################################################################################
#
# Increment the stored serial number and save it.  We are using the format
# YYYYMMDD## for the serial number, which is basically the year, numeric month,
# and numeric day combined with a two digit increment.  This is a common
# format which leaves us with room to make up to 99 changes to DNS throughout
# any single day.
#
################################################################################
sub increment_serial
{
   local($serial,$date);

   $date = &get_date();
   $date .= "00";
   if ($date > $SERIAL)
   {
      $serial = $date + 1;
   }
   else
   {
      $serial = $SERIAL + 1;
   }
   &save_serial($serial);
   return ($serial);
}

################################################################################
#
# Store the serial number in the database.
#
################################################################################
sub save_serial
{
   local($serial) = @_;

   $qs = "UPDATE conf SET serial='$serial' WHERE domainname='$DOMAINNAME'";
   $sth = $dbh->Query($qs)
      or &notification("[save_serial] Query had some problem: $Mysql::db_errstr\n$qs\n");
}
   
################################################################################
#
# Reverse the IP and put it in the standard DNS REV format.
#
################################################################################
sub reverse
{
   local($ip) = @_;

   ($a,$b,$c,$d) = split(/\./, $ip);
   return ("$d.$c.$b.$a.IN-ADDR.ARPA.");
}

################################################################################
#
# Get the DNS header information from the 'conf' database, while incrementing
# the serial number accordingly.
#
################################################################################
sub dns_header
{
   local($forward,$serial) = @_;
   local($qs,$rev,$fwd,$_,@row);

   $qs = "SELECT dns_header FROM conf WHERE domainname='$DOMAINNAME'"; 
   $sth = $dbh->Query($qs)
      or &notification("[dns_header] Query had some problem: $Mysql::db_errstr\n$qs\n");
   @row=$sth->FetchRow;
   $_ = $row[0];
   s/\@SERIAL\@/$serial/g;
   s/
//g;
   $fwd = $_;
   $qs = "SELECT dns_fwd_header FROM conf WHERE domainname='$DOMAINNAME'"; 
   $sth = $dbh->Query($qs)
      or &notification("[dns_header] Query had some problem: $Mysql::db_errstr\n$qs\n");
   @row=$sth->FetchRow;
   $_ = $row[0];
   s/
//g;
   $rev = $_;
   print "$fwd";
   if ($forward)
   {
      print "$rev";
   }
   print "\n";
}

################################################################################
#
# Build the DNS forward and reverse files, pulling the configurable information
# from the 'conf' database.
#
################################################################################
sub build_all_dns
{
   local(%subnets) = @_;

   $serial = &increment_serial();

   open (FWD, "> $TMP/$DOMAINNAME") || die "Could not write out FWD file.\n";
   select(FWD);
   &dns_header(1,$serial);
   select(STDOUT);
   foreach $SUBNET (sort numerically keys %subnets)
   {
      open(REV, ">$TMP/$SUBNET.rev");
      select(REV);
      &dns_header(0,$serial);
      select(STDOUT);
      $qs = "SELECT * FROM name WHERE IP_Address LIKE '$SUBNET.%'";
      $sth = $dbh->Query($qs)
         or &notification("[build_all_dns] Query had some problem: $Mysql::db_errstr\n$qs\n");
      while (@row=$sth->FetchRow)
      {
         ($a,$b,$c,$d) = split(/\./, $row[1]);
         $ip = &padit($a) . &padit($b) . &padit($c) . &padit($d);
         $ip = $ip . ":" . $row[2] . ":" . $row[0];
         $ips{$ip} = 1 if ($ip != 0);
      }
      print FWD ";------------------------------\n";
      print FWD "; Subnet $SUBNET\n";
      print FWD ";------------------------------\n";
      foreach $addr (sort numerically keys %ips)
      {
         ($ip, $name, $id) = split(/:/, $addr);
         $aliases = &get_aliases($id);
         $ip = unpadit($ip);
         $revip = &reverse($ip);
         select (FWD);
         $~ = DNS_A;
         write;
         select(REV);
         $name.= ".$DOMAINNAME.";
         $~ = DNS_REV;
         write;
         select (STDOUT);
         foreach $alias (split(/\,/, $aliases))
         {
            select (FWD);
            $~ = DNS_C;
            write;
            select(STDOUT);
         }
         delete $ips{$addr};
      }
      close (REV);
   }
   close(FWD);
}

################################################################################
#
# Simply parse the database, breaking out subnets, count them all up, and
# generate a page showing how many IP's are used/available for each subnet.
#
################################################################################
sub subnet_usage
{
   local ($used, $left);

   &set_up_page("IP Usage Statistics");
   %subnets=&get_subnets();
   print <<END
<TABLE BORDER=3>
<TR>
<TD ALIGN=CENTER><B>SUBNET</B></TD>
<TD ALIGN=CENTER><B>IP's USED</B></TD>
<TD ALIGN=CENTER><B>IP's AVAILABLE</B></TD>
</TR>
END
;
   foreach $SUBNET (sort numerically keys %subnets)
   {
      $used = 0;
      print "<TR><TD>$SUBNET</TD>";
      $qs="SELECT IP_Address FROM name WHERE IP_Address LIKE '$SUBNET.%'";
      $sth = $dbh->Query($qs)
         or &notification("[subnet_usage] Query had some problem: $Mysql::db_errstr\n$qs\n");
      while (@row=$sth->FetchRow)
      {
         $used ++;
      }
      $left = 256-$used;
      print "<TD ALIGN=CENTER>$used</TD><TD ALIGN=CENTER>$left</TD></TR>";
   }
   print "</TABLE>";
   &end_page(1);
   exit;
}

################################################################################
#
# Call the taintperl script which will actually push the generated files to
# the correct machines.
#
################################################################################
sub push_maps
{
   system("@@SUDO@@ @@ADMIN@@/hostupdate > /tmp/output");
}

################################################################################
#
# Define the numerical sorting routine.
#
################################################################################
sub numerically
{
   $a <=> $b;
}

################################################################################
#
# Bring up the administrative interface
#
################################################################################
sub administer
{
   &set_up_page("Administrative Interface");

   if ($FORM{'UserManager'})
   {
      &user_manager();
      &end_page(1);
      exit;
   }
   if ($FORM{'ConfManager'})
   {
      &conf_manager($FORM{'DOMAINNAME'});
      &end_page(1);
      exit;
   }
   &manager();
   &end_page(1);
}

################################################################################
#
# Show a cleaned-up administrative interface to allow for either management
# of the domain configurations or the user accounts for the system.
#
################################################################################
sub manager
{
   print <<EOF
<TABLE CELLPADDING=0 CELLSPACING=20 BORDER=0>
<FORM METHOD="POST" ACTION="$my_name">
EOF
;
   print "<TH BGCOLOR=\"#aeaeae\"> Domain Management </TH>\n" if ($ACCESS == 0);
   print "<TH BGCOLOR=\"#aeaeae\"> User Management </TH>\n";
   print "<TR>\n";
   if ($ACCESS == 0)
   {
      print "   <TD VALIGN=\"TOP\">\n";
      print "   Modify\n";
      $qs = "SELECT domainname FROM conf";
      $sth = $dbh->Query($qs)
         or &notification("[manager] Query had some problem: $Mysql::db_errstr\n$qs\n");
      print "<SELECT NAME=\"DOMAINNAME\">\n";
      while(@row=$sth->FetchRow)
      {
         print "<OPTION VALUE=\"$row[0]\"> $row[0]\n";
      }
      print <<EOF
</SELECT>
<INPUT TYPE=HIDDEN NAME=\"Admin\" VALUE=\"1\">
<INPUT TYPE=HIDDEN NAME=\"ConfManager\" VALUE=\"1\">
<INPUT TYPE="image" BORDER=0 SRC="@@IMAGES@@/go.gif" ALT="Go!">
</FORM>
   </TD>
EOF
;
   }
   print <<EOF
   <TD VALIGN="TOP">
      <FORM METHOD="POST" ACTION="$my_name">
      <TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>
      <TR>
         <TD VALIGN="TOP">
            Modify
EOF
;
   $qs = "SELECT Username FROM users";
   $qs .= " WHERE Username='$USER'" if ($ACCESS != 0);
   $sth = $dbh->Query($qs)
      or &notification("[manager] Query had some problem: $Mysql::db_errstr\n$qs\n");
   print "<SELECT NAME=\"User_Mod\">\n";
   while(@row=$sth->FetchRow)
   {
      print "<OPTION VALUE=\"$row[0]\"> $row[0]\n";
   }
   print <<EOF
</SELECT>
<INPUT TYPE=HIDDEN NAME=\"Admin\" VALUE=\"1\">
<INPUT TYPE=HIDDEN NAME=\"UserManager\" VALUE=\"1\">
<INPUT TYPE="image" BORDER=0 SRC="@@IMAGES@@/go.gif" ALT="Go!">
</FORM>
         </TD>
      </TR>
EOF
;
   if ($ACCESS == 0)
   {
      print <<EOF
      <TR>
         <TD VALIGN="TOP">
<FORM METHOD="POST" ACTION="$my_name">
Add A New User
<INPUT TYPE=HIDDEN NAME="Admin" VALUE="1">
<INPUT TYPE=HIDDEN NAME="UserManager" VALUE="1">
<INPUT TYPE=HIDDEN NAME="User_Add" VALUE="1">
<INPUT TYPE="image" BORDER=0 SRC="@@IMAGES@@/go.gif" ALT="Go!">
</FORM>
         </TD>
      </TR>
EOF
;
   }
   print <<EOF
      </TABLE>
   </TD>
</TR>
</TABLE>
EOF
;
}

################################################################################
#
# Display the current configuration options as well as the contents of the
# configuration fields.
#
################################################################################
sub conf_manager
{
   local($domainname) = @_;

   $qs = "SELECT dns_header,dns_fwd_header,yp_header,dhcpd_header,named_restart,yp_restart,dhcpd_restart FROM conf WHERE domainname='$domainname'";
   $sth = $dbh->Query($qs)
      or &notification("[conf_manager] Query had some problem: $Mysql::db_errstr\n$qs\n");
   @row=$sth->FetchRow;
   $dns_header = $row[0];
   $dns_fwd_header = $row[1];
   $nis_header = $row[2];
   $dhcpd_header = $row[3];
   $DNS_RESTART = $row[4];
   $YP_RESTART = $row[5];
   $DHCPD_RESTART = $row[6];
   print <<EOF
<FORM METHOD="POST" ACTION="@@CGIBIN@@/$my_name">
<INPUT TYPE=HIDDEN NAME="UPDATE_CONF" VALUE="1">
<INPUT TYPE=HIDDEN NAME="DOMAINNAME" VALUE="$DOMAINNAME">
<TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0 WIDTH=100%>
<TR>
   <TD>
      <TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>
      <TH COLSPAN=2 BGCOLOR="#aeaeae"> Global Settings</TH>
      <TR>
         <TD>Temporary Directory : </TD>
         <TD><INPUT TYPE=TEXT NAME="TMP" VALUE="$TMP" SIZE=20></TD>
      </TR>
      <TR>
         <TD>IP Start Range : </TD>
         <TD><INPUT TYPE=TEXT NAME="RANGE1" VALUE="$RANGE[0]" SIZE=3></TD>
      </TR>
      <TR>
         <TD>IP End Range : </TD>
         <TD><INPUT TYPE=TEXT NAME="RANGE2" VALUE="$RANGE[1]" SIZE=3></TD>
      </TR>
      </TABLE>
   </TD>
   <TD>
      Global Switches
      <TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>
      <TH COLSPAN=2 BGCOLOR="#aeaeae"> Domain Name: $DOMAINNAME</TH>
      <TR>
         <TD>Manage NAMED/BIND : </TD>
         <TD><INPUT TYPE=CHECKBOX NAME="NAMED_ACTIVE"
EOF
;
   print " CHECKED" if ($NAMED_DIR);
   print <<EOF
></TD>
      </TR>
      <TR>
         <TD>Manage NIS/YP : </TD>
         <TD><INPUT TYPE=CHECKBOX NAME="YP_ACTIVE"
EOF
;
   print " CHECKED" if ($NIS_DIR);
   print <<EOF
></TD>
      </TR>
      <TR>
         <TD>Manage DHCPD : </TD>
         <TD><INPUT TYPE=CHECKBOX NAME="DHCPD_ACTIVE"
EOF
;
   print " CHECKED" if ($DHCPD_DIR);
   print <<EOF
></TD>
      </TR>
      </TABLE>
   </TD>
</TR>
<TR>
   <TD COLSPAN=2>
      <TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>
      <TH COLSPAN=2 BGCOLOR="#aeaeae"> NAMED/BIND Configuration</TH>
      <TR>
         <TD>File Locations : </TD>
         <TD><INPUT TYPE=TEXT NAME="NAMED_DIR" VALUE="$NAMED_DIR" SIZE=20></TD>
      </TR>
      <TR>
         <TD>Header for Forward File : </TD>
         <TD><TEXTAREA NAME="dns_header" ROWS=5 COLS=80>$dns_header</TEXTAREA></TD>
      </TR>
      <TR>
         <TD>Header for Reverse File : </TD>
         <TD><TEXTAREA NAME="dns_fwd_header" ROWS=5 COLS=80>$dns_fwd_header</TEXTAREA></TD>
      </TR>
      <TR>
         <TD>Restart/Reload Command : </TD>
         <TD><INPUT TYPE=TEXT NAME="DNS_RESTART" VALUE='$DNS_RESTART' SIZE=30></TD>
      </TR>
      </TABLE>
   </TD>
</TR>
<TR>
   <TD COLSPAN=2>
      <TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>
      <TH COLSPAN=2 BGCOLOR="#aeaeae"> NIS/YP Configuration</TH>
      <TR>
         <TD>File Locations : </TD>
         <TD><INPUT TYPE=TEXT NAME="NIS_DIR" VALUE="$NIS_DIR" SIZE=20></TD>
      </TR>
      <TR>
         <TD>Header for Files : </TD>
         <TD><TEXTAREA NAME="nis_header" ROWS=5 COLS=80>$nis_header</TEXTAREA></TD>
      </TR>
      <TR>
         <TD>Restart/Reload Command : </TD>
         <TD><INPUT TYPE=TEXT NAME="YP_RESTART" VALUE='$YP_RESTART' SIZE=30></TD>
      </TR>
      </TABLE>
   </TD>
</TR>
<TR>
   <TD COLSPAN=2>
      <TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>
      <TH COLSPAN=2 BGCOLOR="#aeaeae"> DHCPD Configuration</TH>
      <TR>
         <TD>File Locations : </TD>
         <TD><INPUT TYPE=TEXT NAME="DHCPD_DIR" VALUE="$DHCPD_DIR" SIZE=20></TD>
      </TR>
      <TR>
         <TD>Header for Files : </TD>
         <TD><TEXTAREA NAME="dhcpd_header" ROWS=5 COLS=80>$dhcpd_header</TEXTAREA></TD>
      </TR>
      <TR>
         <TD>Restart/Reload Command : </TD>
         <TD><INPUT TYPE=TEXT NAME="DHCPD_RESTART" VALUE='$DHCPD_RESTART' SIZE=30></TD>
      </TR>
      </TABLE>
      <HR SIZE=1 NOSHADE>
      <TABLE>
      <TR>
         <TD>
            Delete OS: <SELECT NAME="OS_Del">
            <OPTION VALUE="0"> None
EOF
;
   $qs = "SELECT id,name FROM OS";
   $sth = $dbh->Query($qs)
      or &notification("[conf_manager] Query had some problem: $Mysql::db_errstr\n$qs\n");
   while (@row=$sth->FetchRow)
   {
      print "<OPTION VALUE=\"" . $row[0] . "\"> " . $row[1] . "\n";
   }
   print <<EOF
            </SELECT>
         </TD>
         <TD>
            Add OS: <INPUT TYPE=text NAME="OS_Add" SIZE=15>
         </TD>
      </TR>
      </TABLE>
      <HR SIZE=1 NOSHADE>
      <TABLE>
      <TR>
         <TD>
            Delete Processor Type: <SELECT NAME="Processor_Del">
            <OPTION VALUE="0"> None
EOF
;
   $qs = "SELECT id,name FROM processor_types";
   $sth = $dbh->Query($qs)
      or &notification("[conf_manager] Query had some problem: $Mysql::db_errstr\n$qs\n");
   while (@row=$sth->FetchRow)
   {
      print "<OPTION VALUE=\"" . $row[0] . "\"> " . $row[1] . "\n";
   }
   print <<EOF
            </SELECT>
         </TD>
         <TD>
            Add Processor Type: <INPUT TYPE=text NAME="Processor_Add" SIZE=15>
         </TD>
      </TR>
      </TABLE>
      <HR SIZE=1 NOSHADE>
   </TD>
</TR>
<TR>
   <TD ALIGN=CENTER>
      <TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>
      <TR>
         <TD>
            <INPUT TYPE=SUBMIT NAME="Update" VALUE="Save Changes">
         </TD>
         <TD>
            <INPUT TYPE=SUBMIT NAME="Update" VALUE="Cancel">
         </TD>
      </TR>
      </TABLE>
   </TD>
</TR>
</TABLE>
</FORM>
EOF
;
}

################################################################################
#
# Diplay a semi-clean interface to allow the end-user to update/administer
# the user accounts for this system.
#
################################################################################
sub user_manager
{
   local($i,$qs,@fields,@field_names) = 0;

   $qs = "SHOW FIELDS FROM users";
   $sth = $dbh->Query($qs)
      or &notification("[user_manager] Query had some problem: $Mysql::db_errstr\n$qs\n");
   while (@row=$sth->FetchRow)
   {
      $fields[$i] = $row[0];
      $i++;
   }
   if (! $FORM{'User_Add'})
   {
      $qs = "SELECT Username,Password,uid,Email,Access FROM users WHERE Username='$FORM{'User_Mod'}'";
      $sth = $dbh->Query($qs)
         or &notification("[user_manager] Query had some problem: $Mysql::db_errstr\n$qs\n");
      @row=$sth->FetchRow;
   }
   print <<EOF
<FORM METHOD="POST" ACTION="$my_name">
<INPUT TYPE="HIDDEN" NAME="uid" VALUE="$row[2]">
<TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>
EOF
;
   for ($i=0; $i<=$#fields; $i++)
   {
      next if ($fields[$i] eq "uid");
      print "<TR>\n";
      print "   <TD>\n";
      print "      $fields[$i] :\n";
      print "   </TD>\n";
      print "   <TD>\n";
      if ($fields[$i] eq "Access")
      {
         if ($USER eq $row[0])
         {
            $qs = "SELECT description FROM access WHERE id='$row[$i]'";
            $sth = $dbh->Query($qs)
               or &notification("[user_manager] Query had some problem: $Mysql::db_errstr\n$qs\n");
            @access_levels=$sth->FetchRow;
            print "$access_levels[0]";
         }
         else
         {
            print "<SELECT NAME=\"$fields[$i]\">\n";
            $qs = "SELECT id,description FROM access";
            $sth = $dbh->Query($qs)
               or &notification("[user_manager] Query had some problem: $Mysql::db_errstr\n$qs\n");
            while (@access_levels=$sth->FetchRow)
            {
               next if ($ACCESS > $access_levels[0]);
               print "<OPTION VALUE=\"$access_levels[0]\"";
               print " SELECTED" if ($row[$i] eq $access_levels[0]);
               print "> $access_levels[1]\n";
            }
            print "</SELECT>\n";
         }
      }
      else
      {
         print "      <INPUT TYPE=\"TEXT\" NAME=\"$fields[$i]\" SIZE=\"20\"";
         print " VALUE=\"$row[$i]\"" if ( (! $FORM{'User_Add'}) && ($fields[$i] ne "Password") );
         print "><BR>\n";
      }
      print "   </TD>\n";
      print "   <TD>\n";
      print "      <A HREF=\"javascript://\" onClick=\"openDescription('$fields[$i]');\"><img src=\"@@IMAGES@@/descriptionButton.gif\" width=\"10\" height=\"10\" border=\"0\" alt=\"Click here for more information about the $fields[$i] field\"></A>\n";
      print "   </TD>\n";
      print "</TR>\n";
   }
   if ($FORM{'User_Add'})
   {
      print <<EOF
<TR>
   <TD>
      <INPUT TYPE="SUBMIT" NAME="User_Mod" VALUE="Add">
   </TD>
</TR>
EOF
;
   }
   else
   {
      print <<EOF
<TR>
   <TD>
      <INPUT TYPE="SUBMIT" NAME="User_Mod" VALUE="Modify">
   </TD>
   <TD>
      <INPUT TYPE="SUBMIT" NAME="User_Mod" VALUE="Delete">
   </TD>
</TR>
EOF
;
   }
   print "</TABLE>\n";
   print "</FORM>\n";
}

################################################################################
#
# A small crypt function to simulate the standard UNIX crypt routine.  This
# should insure that the password is happily encrypted.
#
################################################################################
sub crypt_passwd
{
   local($in) = @_;
   local($out,$seed,@possible);

   @possible = {'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z'};
   $seed = $possible[rand($#possible)] . $possible[rand($#possible)];
   $out = crypt($in, $seed);
   return ($out);
}

################################################################################
#
# A quick routine to allow for updating the user information, including
# the password.
#
################################################################################
sub modify_user
{
   local($crypt);

   $crypt = &crypt_passwd($FORM{'Password'}) if ($FORM{'Password'});
   if ($FORM{'User_Mod'} eq "Add")
   {
      $qs = "INSERT INTO users VALUES ('$FORM{'Username'}', '$crypt', '', '$FORM{'Email'}', '$FORM{'Access'}')";
   }
   if ($FORM{'User_Mod'} eq "Modify")
   {
      $qs = "UPDATE users SET Username='$FORM{'Username'}',Email='$FORM{'Email'}'";
      $qs .= ",Access='$FORM{'Access'}'" if ($FORM{'Access'});
      $qs .= ",Password='$crypt'" if ($FORM{'Password'});
      $qs .= " WHERE uid='$FORM{'uid'}'";
   }
   if ($FORM{'User_Mod'} eq "Delete")
   {
      $qs = "DELETE FROM users WHERE uid='$FORM{'uid'}'";
   }
   $sth = $dbh->Query($qs)
      or &notification("[user_manager] Query had some problem: $Mysql::db_errstr\n$qs\n");
}

################################################################################
#
# Update the configuration database
#
################################################################################
sub update_config
{
   $FORM{'NAMED_DIR'} = "" if (! $FORM{'NAMED_ACTIVE'});
   $FORM{'NIS_DIR'} = "" if (! $FORM{'YP_ACTIVE'});
   $FORM{'DHCPD_DIR'} = "" if (! $FORM{'DHCPD_ACTIVE'});
   $qs="UPDATE conf set named='$FORM{'NAMED_DIR'}',yp='$FORM{'NIS_DIR'}',dhcpd='$FORM{'DHCPD_DIR'}',tmp='$FORM{'TMP'}',start_range='$FORM{'RANGE1'}',end_range='$FORM{'RANGE2'}',dns_header='$FORM{'dns_header'}',dns_fwd_header='$FORM{'dns_fwd_header'}',yp_header='$FORM{'nis_header'}',dhcpd_header='$FORM{'dhcpd_header'}',named_restart='$FORM{'DNS_RESTART'}',yp_restart='$FORM{'YP_RESTART'}',dhcpd_restart='$FORM{'DHCPD_RESTART'}' WHERE domainname='$FORM{'DOMAINNAME'}'";
   $sth = $dbh->Query($qs)
      or &notification("[update_config] Query had some problem: $Mysql::db_errstr\n$qs\n");
   if ($FORM{'OS_Add'})
   {
      $qs="INSERT INTO OS VALUES ('', '$FORM{'OS_Add'}')";
      $sth = $dbh->Query($qs)
         or &notification("[update_config] Query had some problem: $Mysql::db_errstr\n$qs\n");
   }
   if ($FORM{'OS_Del'})
   {
      $qs="DELETE FROM OS WHERE id=$FORM{'OS_Del'}";
      $sth = $dbh->Query($qs)
         or &notification("[update_config] Query had some problem: $Mysql::db_errstr\n$qs\n");
   }
   if ($FORM{'Processor_Add'})
   {
      $qs="INSERT INTO processor_types VALUES ('', '$FORM{'Processor_Add'}')";
      $sth = $dbh->Query($qs)
         or &notification("[update_config] Query had some problem: $Mysql::db_errstr\n$qs\n");
   }
   if ($FORM{'Processor_Del'})
   {
      $qs="DELETE FROM processor_types WHERE id=$FORM{'Processor_Del'}";
      $sth = $dbh->Query($qs)
         or &notification("[update_config] Query had some problem: $Mysql::db_errstr\n$qs\n");
   }
}

################################################################################
#
# Show a standard error screen if we happen to hit any odd errors when opening
# or writing to any files.
#
################################################################################
sub error_out
{
   local($error) = @_;

   &set_up_page("Unexpected Error!");
   print <<END
There has been an unexpected server-side error.  Please contact the
Administrator to determine the cause of the error.<BR>
<BR>
<B>ERROR: $error</B>
END
;
   &end_page(1);
   exit;
}

################################################################################
#
# Do a basic check to make sure the entry is not null.
# (This is to be replaced with a more thorough check which really looks
# (at the IP, hostname, and aliases to make sure there is no error there.)
#
################################################################################
sub check_entry
{
   local(@fields) = @_;

   foreach $count (0..$#fields)
   {
      if ($FORM{$field[$count]} eq "")
      {
         $error = 1;
      }
   }
   return ($error);
}

################################################################################
#
# Display a status notice after each change to the database.
#
################################################################################
sub notification
{
   local($notice) = @_;

   SWITCH:
   {
      if ($notice eq 1)
      {
         print <<END
<B><BLINK>ERROR</BLINK></B>: No such entry, "
END
;
         print "$FORM{'Query'}" if ($FORM{'Query'});
         print "$FORM{'Delete'}" if ($FORM{'Delete'});
         print "$FORM{'Modify'}" if ($FORM{'Modify'});
         print <<END
", found in the database under "$FORM{'query_type'}".  Please re-submit.<BR>
END
;
         last SWITCH;
      }
      if ($notice eq 2)
      {
         
         print "<B><BLINK>NOTICE</BLINK></B>: ";
         print "$FORM{'Machine_Name'} has been <B>deleted</B> from the database.<BR>" if ($FORM{'submit'} eq "delete");
         print "$FORM{'Machine_Name'} has been updated.<BR>" if ($FORM{'submit'} eq "modify");
         print "$FORM{'Machine_Name'} has been added with the IP address of <B>$FORM{'IP_Address'}</B><BR>" if ($FORM{'submit'} eq "add");
         print "$FORM{'Machine_Name'} has been moved to the IP address of <B>$FORM{'IP_Address'}</B><BR>" if ($FORM{'function'} eq "Move");
         last SWITCH;
      }
      if ($notice eq 3)
      {
         print "<B><BLINK>NOTICE</BLINK></B>: ";
         print "The maps have been <B><I>PUSHED</I></B>.<BR>";
         last SWITCH;
      }
      if ($notice eq 4)
      {
         print "<B><BLINK>NOTICE</BLINK></B>: ";
         print "The configuration settings have been <B><I>UPDATED</I></B>.<BR>";
         last SWITCH;
      }
      if ($notice eq 5)
      {
         print "<B><BLINK>NOTICE</BLINK></B>: ";
         print "The configuration changes have been <B><I>Cancelled</I></B>. No Changes have been made.<BR>";
         last SWITCH;
      }
      if ($notice eq 6)
      {
         print "<B><BLINK>NOTICE</BLINK></B>: ";
         print "The account changes for the user \"$FORM{'Username'}\" have been <B><I>Completed</I></B>.";
         last SWITCH;
      }
      if ($notice eq 20)
      {
         print "<B><FONT COLOR=\"RED\"><BLINK>ERROR</BLINK>:</FONT></B> ";
         print "That machine name is already in use.<BR>";
         last SWITCH;
      }
      if ($notice eq 21)
      {
         print "<B><FONT COLOR=\"RED\"><BLINK>ERROR</BLINK>:</FONT></B> ";
         print "That IP address is already in use.<BR>";
         last SWITCH;
      }
      if ($notice eq 22)
      {
         print "<B><FONT COLOR=\"RED\"><BLINK>ERROR</BLINK>:</FONT></B> ";
         print "That hardware address is already in use.<BR>";
         last SWITCH;
      }
      if ($notice eq 23)
      {
         print "<B><FONT COLOR=\"RED\"><BLINK>ERROR</BLINK>:</FONT></B> ";
         print "No machine name or IP address was specified.<BR>";
         last SWITCH;
      }
      print "<B><FONT COLOR=\"RED\"><BLINK>ERROR</BLINK>:</FONT></B> ";
      print "$notice.<BR>";
   } 
}

################################################################################
#
# Give some help information.
#
################################################################################
sub help
{
   local($describe) = @_;
   local($qs);

   $qs = "SELECT info FROM descriptions WHERE item='$describe'";
   $sth = $dbh->Query($qs)
      or &notification("[help] Query had some problem: $Mysql::db_errstr\n$qs\n");
   @row=$sth->FetchRow;
   print "$row[0]";
}

################################################################################
#
# Check to see if the item exists in the database already.
#
################################################################################
sub exists
{
   local($item,$field) = @_;
   local($sth,$qs,$exists,@row);

   $qs = "SELECT * FROM name WHERE $field='$item'";
   $qs .= " AND id != '$FORM{'id'}'" if ($FORM{'submit'} ne "add" );
   $sth = $dbh->Query($qs)
      or &notification("[exists] Query had some problem: $Mysql::db_errstr\n$qs\n");
   @row=$sth->FetchRow;
   $exists = 1 if ($row[0]);
   return ($exists);
}

################################################################################
#
# Display the change log on-line rather than forcing the user to search for
# and read a particular file.
#
################################################################################
sub show_changelog
{
   $changes = <<EOF
1.13
- Minor update to the list of requirements to include the DBI perl module.
- Fixed a bug in the creation of the ethers file.
- Fixed a bug in the update of the aliases.  There was a typo in the SQL line
  which prevented the old aliases from being removed.
- Migrated away from the master 'root' DB user to a singular user with
  restricted access to managing the machine database ONLY.
- Implemented the idea of moving a host from one subnet to another.

1.12
- Fixed a show stopper bug in the error/sanity checking which prevented updating
  any of the existing records.
- Added the mk.netgroup script to the available utils.
- Updated the makefile to support the ability to make the utilities.
- Updated the UPGRADE doc.
- Fixed the display of the notice when deleting a record from the database.
- Cleaned up some more comments for the new functions which handle the display
  of the changes to the records.
- Very minor bug fixes in the display of the record via a query or delete.

1.11
- Created and implemented a true modify function.  Deleting the entry and
  re-adding it, while saving code space, meant that each change would
  increment the id number.  Many changes meant that the system would be eating
  a heck of a lot of id numbers for the sake of saving about 10 lines of code.
- Minor change: cleaned up the "query_string" vs "qs" issue.  All query_string
  variables have been replaced by qs.
- Fixed/Changed the Modify and Delete forms to send the id number rather than
  the query_type and the contents of the query.  This fixed situations where
  an error would result if the original query was of something outside of the
  "name" table.
- Changed the "notify" function to be more aptly named "notification".
- Added some functionality to the "notification" function and moved all of the
  "warn" outputs from the queries into it.  This way, any warnings will go to
  the web interface rather than into the apache log file.
- Fixed all of the old "warn" messages to include the correct function in the
  messaage.
- Removed the useless "error" function.
- Introduced different user levels: Display only, Modify, Administer
- Created "access" table for the different user levels.
- Updated the user management form to reflect the new user levels.
- Fixed the cookie code to actually retain the user info as opposed to simply
  creating the cookie then wiping out the actual value.
- Splashed the user level access control restrictions throughout the entire
  codebase.
- Only Admins can create new accounts and modify the access levels of existing
  accounts.
- Added new fields to the "specs" table.
- Cleaned up the usage of the "noshow" flag.
- Updated comments within the code.
- Added help info for the new fields.
- Cleaned up the "entry" variable, making it a true array.  This removes
  all of the remaining cruft from the original design.  I was forced to make
  it a global because I can not figure out how to pass two arrays to a 
  function and have them remain intact.
- Updated query, add, remove, modify, and search_db to compensate for the
  new fields and the correction to the array.
- Updated the dump_database routines to handle the new tables and entries.
- Created a "changes" table to keep some rudimentary "last updated" information.
- Added the information from the "changes" table to the display output.
- Updated all of the database fields for easier readability.
- Put small hooks into each of the displays to remove 'pretty up' the display
  of the field names.
- Created a new upgrade script to handle upgrading from v1.10 to v1.11.  This
  particular upgrade has a whole mess of table additions, field additions to
  existing tables, and field name updates.
- Updated the "make upgrade" install process and updated the UPGRADE doc.
- Created the AUTHOR and CONTRIB files.
- Fixed a minor bug in the notification generated when adding a new host.
- Added the hosts2sql.pl script as a helpful utility.

1.10
- Broke apart the administrative interface into sub-pages to allow for
- Added option to administrative config menu to add/delete items from
  the OS listing.
- Re-built the intial administrative interface to give all of the new options.
- Built the pages for the modify/delete and add of users.
- Re-worked the code for the domain configuration and user maintenance to
  work with the new page layout.
- Moved the serial number OUT of the DNS files and into the conf table.  I
  do not know why I didn't do this before, but, without it, it certainly makes
  it painful to recover everything in the event of an emergency.
- DNS serial number is now based upon the date + a sequence number.
  the addition of the user management pages.
- Created function to crypt password before entering it to the database.
- Fixed minor typo in the creation of the yp_header.
- Moved the OS definition out of the "name" table and into its own table.
  I plan on expanding this table soon to include items like Serial Number, etc.
- Fixed a hidden bug in the display SELECT mysql call which exposed itself when
  the "specs" table was created.
- Created the upgrade script and the corresponding Makefile option.
- Created the UPGRADE doc.
- Made the examples directory and threw in some example header files.

1.9
- Added in the javascript to handle the "close this window" functionality for
  the help pop-ups.
- Fixed the cookie update.  Now, the cookie will 'refresh' with each use, and
  providing that the cookie hasn't expired, it will update the life of the
  cookie.  This will reduce unnecessary re-logins.
- Added header links to the display section when displaying the contents of
  the database.  This allows for on-the-fly resorting of the data.
- Added Modify and Delete buttons to the Query output page.
- Fixed a typo in the describe table for the alias help.
- Fixed some typos in the README and updated the requirements slightly.
- Fixed a configuration problem where I left a bunch of /cgi-bin/ items
  hardcoded to that location, rather than taking the value defined in the
  Makefile.
- Cleaned up the output of the display function (used for displaying the
  results of the query, modify, and delete options), but it definitely made
  the code a little nastier.  The output is now much friendlier on the eyes, 
  but still lacking any real color.
- Added an IP entry to the descriptions table.
- Fixed a typo in the javascript pop-up help function which show up under IE.

1.8
- Created the Makefile
- Changed the source directory structure for easier installation/upgrade
- Replaced all key variables in the source files with flags which get
  replaced upon installation/upgrade via the Makefile.
- Created the README and INSTALL text files.
- Updated the hostupdate script to pull all of its configuration information
  out of the database.
- Added in the restart command sequences for named, yp, and dhcpd to make it
  fully configurable.
- Created the administrative interface layout.
- Properly populated the layout.
- The Administrative interface now updates the configuration file.
- Added the proper notification messages for updating and cancelling the updates
  to the configuration database.
- Put in hooks for global on/off switches for DNS, DHCP, and YP on the admin
  page and used them to set/clear the directory setting for the particular
  service.  The program checks to see if that directory setting is set to
  anything to determine if we should process the files for that service.
- Fixed a minor typo in the error listing.  I accidentally listed error#21
  twice and left off error#22.
- Created error#23 for situations when a hostname or IP is not specified.  We
  cannot have blank entries in the resulting files.

1.7
- Added in the description graphic.
- Added in the javascript to handle the pop-ups.
- Created the "descriptions" table in the database.
- Added the "help" routine to handle when the description graphic is clicked on.
- Populated the "descriptions" table.
- Removed the two "NOTE" lines because of the new help format.
- Added in some rudimentary error checking when adding a new host.  It checks
  the hostname, IP, and hardware address to see if any of them are in use.  If
  they are, it generates the appropriate error message.
- Added the configuration aspect for DHCP.
- Added in the "comment" and "OS" fields.  "OS" is a little special because of
  the pull-down selection capability.
- Fixed a bug in the display by IP function.  It was something I fixed some
  time ago in the display by other function, but, oddly, I never made the same
  change.  Hence, things broke in funny ways when I added the two new fields.
- Started creating the Administration interface.  It barely shows anything and
  you cannot do anything with it yet.

1.6
- Added comments to the uncommented sections of the code.
- Updated the initial description at the top of the script.
- Broke apart the dump_database routine.  Hopefully, this will make
  it easier to build the sort functions for the other columns.
- Sorting now sort-of works on the other columns (except aliases).  I need to
  work on the SQL query a bit more because of the way the join is working.
- Querying by ownername and location now works.  Keep in mind that this will
  only result in the first record that matches.
- New Graphic.  The program is a bit more than just a DNS database now..
- Created a main table for all of the options.
- Added in the creation of the ethers NIS map.  This was specifically done
  to support the use of Solaris's jumpstart.
- Added logout capability.
- Code now always adds to all of the database fields, even if there is no
  content.  This makes searching a whole heck of a lot easier.
- Redid the login screen.  Created a graphical "badge" that needs to be
  filled in.
- Created a hook for the administrative interface, but I have not built it yet.
- Fixed a bug in the display of the Entire database when sorting by IP where
  it would keep adding to the entries to be displayed.  Not that it existed
  when displaying other items, I reflected the bug change there, as well, in
  the event that we change something down the road.
- Fixed the display of the "log out" option at the bottom of the screen and
  increased the delay of the "Processing Log Out..." option to allow for
  the cookie to be completely destroyed and for the browser to notice its
  removal.

1.5
- Databased all of the configuration information.  This will build a foundation
  for a web-based configuration page.
- Databased all of the file header information.  This, too, will allow for
  a web-based customization page.
- Hacked my script to handle the restarting of the dhcp daemon, which runs on
  another server.
- Added this on-line changelog routine. :)

1.4 (never in production)
- Cleaned up the DHCPd configuration options.  It should set the NIS domain
  name amongst other information.
- Fixed the DNS TTL flag.
- Other changes that I forgot about over the Xmas break.

1.3
- Cleaned up the interface.  Folded a lot of the original interface into one
  pull-down form.
- Fixed the sort routine for the IP's.  Basically, I had to break apart the
  IP address on the fly and sort it that way.  A bit expensive, but it works.
  Sorting still only works for the first three fields. :(
- Wrote the DHCPd routines.  Restarting the dhcp daemon by hand is still
  necessary.
- Created the cookie interface to the script.

1.2
- Added in some error checking.  Certain people were using invalid DNS
  characters for hostnames, despite my pleas.  I dropped them all into an
  array, and simply remove them from the hostname.  I should fully fill out
  the array soon.
- Added some (read that as "very minimal") on-line documentation.
- Created a graphic for the top of the page.
- Since people were confused by the lack of a "submit" button, I created
  a "go" graphic and put that into the form.
- Cleaned up the file output for the DNS and NIS services.
- Installed sudo.
- Moved the ownership of the scripts and such over to the "janitor" account,
  and wrote a script, to which the janitor account has access to execute, which
  moves all of the files into place.
- Re-did the login screen.

1.1
- Query/sort via IP, machinename, or hardware adress.
- Added a little error checking, but not much.
- Turned on DNS and NIS map creation and rolled it into production.
- Minor bug fixes per request.
- Migrated the login information over to a database.

1.0
- Migration from the old code that I wrote for Sarnoff.
- Got it working under the Diva setup.
- Modified to make it a little more portable.
- Migrated from a flat-file database to a MySQL database.
EOF
;
   print "<PRE>\n";
   print "$changes\n";
   print "</PRE>\n";
}
